// MTIDE.CPP
/**************************************************************************/
/**                                                                      **/
/**                  LIBRAIRIE DE CALCUL DES MAREES                      **/
/**                                                                      **/
/**************************************************************************/

#include <windows.h> // Of course : temporaire...

#include "mtide.h"
#include "fourier.h"



// TIDE : Renvoie la hauteur de la maree pour un instant t
LDOUBLE tide (LDOUBLE t,MINT onde)
{
	MINT i;
	LDOUBLE result;
	result = Z0;
	for (i=onde;i<NumOnde;i++)
	{
		result+=ondes[i].ampl*cos(ondes[i].speed*t-ondes[i].etab);
	}
	return (result);
}

// DTIDE : Renvoie la derivee de la hauteur d'eau a un instant t
LDOUBLE dtide (LDOUBLE t, MINT onde)
{
	MINT i;
	LDOUBLE result;
	result = 0.0;
	for (i=onde;i<NumOnde;i++)
	{
		result+=-ondes[i].speed*ondes[i].ampl*sin(ondes[i].speed*t-ondes[i].etab);
	}
	return (result);
}

// DDTIDE : Renvoie la derivee seconde de la hauteur d'eau a un instant t
LDOUBLE ddtide (LDOUBLE t, MINT onde)
{
	MINT i;
	LDOUBLE result;
	result = 0.0;
	for (i=onde;i<NumOnde;i++)
	{
		result+=-ondes[i].speed*ondes[i].speed*ondes[i].ampl*cos(ondes[i].speed*t-ondes[i].etab);
	}
	return (result);
}

// NEWTON : Recherche de maximum de la fonction
LDOUBLE newton (LDOUBLE fonction (LDOUBLE),
				LDOUBLE premiere (LDOUBLE),
				LDOUBLE seconde (LDOUBLE),
				LDOUBLE poMINT,
				LDOUBLE precision)
				// newton (x) = x-premiere(x)/seconde(x)		 Suite de Newton
{
	LDOUBLE p,pp;
	DOUBLE dp;
	LLONG n, nmax = 30;
	n = 0;
	p = poMINT-premiere(poMINT)/seconde(poMINT);
	while (p < 0) { p = fabs(modf(p/dop,&dp))*dop; }
	while (p > 2*dop) { p = modf(p/dop,&dp)*dop; }
	pp = p-premiere(p)/seconde(p);
	while (pp < 0) { pp = fabs(modf(pp/dop,&dp))*dop; }
	while (pp > 2*dop) { pp = modf(pp/dop,&dp)*dop; }
	while ((n<=nmax) && (((pp<0) || ((fonction(pp))<0)) ||
		(fabs ((float)(fonction(pp)-fonction(p)))>precision) // Special
		/*(fabs((float)(pp-p))>precision)*/ ))
	{
		n++;
		p = pp;
		pp = p-premiere(p)/seconde(p);
		while (pp < 0) { pp = fabs(modf(pp/dop,&dp))*dop; }		//Special
		while (pp > 2*dop) { pp = modf(pp/dop,&dp)*dop; }		//Special
		if ((fonction (pp)) < 0) { pp += dop/2; }	  	// Special
	}
	//if (n>nmax) { fprMINTf (flog,"...Dpassement dans extremums de Newton...\n");}
	return (pp);
}

// GET_TIDE : Renvoie la maree enregistree.
LDOUBLE get_tide (LDOUBLE t)		// t : Temps en secondes
{
	LDOUBLE gta,gtb,gtc,res;
	gta = tab[(LLONG)(t/PRECISION)];
	gtb = tab[(LLONG)(t/PRECISION)+1];
	gtc = t-((LLONG)(t/PRECISION))*PRECISION;
	res = ((gta*(PRECISION-gtc)+gtb*gtc)/PRECISION);
	return ((res/100));
}

// GET_DTIDE : Renvoie la derivee de la maree enregistree.
LDOUBLE get_dtide (LDOUBLE t)		// t : Temps en secondes
{
	LDOUBLE gta,gtb,gtc,res;
	gta = tab[(LLONG)(t/PRECISION)];
	gtb = tab[(LLONG)(t/PRECISION)+1];
	res = ((gtb-gta)/PRECISION);
	return ((res/100));
}

// GET_DDTIDE : Renvoie la derivee seconde de la maree enregistree.
LDOUBLE get_ddtide (LDOUBLE t)		// t : Temps en secondes
{
	LDOUBLE gta,gtb,gtc,res;
	gta = tab[(LLONG)(t/PRECISION)-1];
	gtb = tab[(LLONG)(t/PRECISION)];
	gtc = tab[(LLONG)(t/PRECISION)+1];
	res = (((gtc-gtb)/PRECISION-(gtb-gta)/PRECISION)/PRECISION);
	return ((res/100));
}

// RGET_TIDE : Renvoie la maree rectifiee.
LDOUBLE rget_tide (LLONG t)		// t : Temps en secondes
{
	LDOUBLE gta,gtb,gtc,res;
	gta = tab[(LLONG)(t/PRECISION)];
	gtb = tab[(LLONG)(t/PRECISION)+1];
	gtc = t-((LLONG)(t/PRECISION))*PRECISION;
	res = ((gta*(PRECISION-gtc)+gtb*gtc)/PRECISION);
	return ((res/100)-tide(t));
}

// RGET_DTIDE : Renvoie la derivee de la maree rectifiee.
LDOUBLE rget_dtide (LLONG t)		// t : Temps en secondes
{
	LDOUBLE gta,gtb,gtc,res;
	gta = tab[(LLONG)(t/PRECISION)];
	gtb = tab[(LLONG)(t/PRECISION)+1];
	res = ((gtb-gta)/PRECISION);
	return ((res/100)-dtide(t));
}

// RGET_DDTIDE : Renvoie la derivee seconde de la maree rectifiee.
LDOUBLE rget_ddtide (LLONG t)		// t : Temps en secondes
{
	LDOUBLE gta,gtb,gtc,res;
	gta = tab[(LLONG)(t/PRECISION)-1];
	gtb = tab[(LLONG)(t/PRECISION)];
	gtc = tab[(LLONG)(t/PRECISION)+1];
	res = (((gtc-gtb)/PRECISION-(gtb-gta)/PRECISION)/PRECISION);
	return ((res/100)-ddtide(t));
}

// DET_ONDE : Determine l'onde de periode p
LDOUBLE det_onde_fonction (LDOUBLE x)
{
	LDOUBLE h, resu;
	LLONG n;
	n = 0;
	resu = 0.0;
	if (x>=0)
	{
		for (h=x;h<(LDOUBLE)(NUMJOURS-2)*24.*3600.;h+=dop)
		{
			resu += rget_tide (h);
			n++;
		}
		resu/=n;
	}
	return (resu);
}

LDOUBLE det_onde_premiere (LDOUBLE x)
{
	LDOUBLE h, resu;
	LLONG n;
	n = 0;
	resu = 0.0;
	if (x>=0)
	{
		for (h=x;h<(LDOUBLE)(NUMJOURS-2)*24.*3600.;h+=dop)
		{
			resu += rget_dtide (h);
			n++;
		}
		resu/=n;
	}
	return (resu);
}

LDOUBLE det_onde_seconde (LDOUBLE x)
{
	LDOUBLE h, resu;
	LLONG n;
	n = 0;
	resu = 0.0;
	if (x>=0)
	{
		for (h=x;h<(NUMJOURS-2)*24.*3600.;h+=dop)
		{
			resu += rget_ddtide (h);
			n++;
		}
		resu/=n;
		return (resu);
	}
	else
	{
		return (1.0);
	}
}

struct tonde det_onde_periode_fourier (LDOUBLE p);

struct tonde det_onde_periode (LDOUBLE p)
{
	LDOUBLE etab,ampl;
	struct tonde ret;
	dop = p;
	if (CFourier == FALSE)
	{
		etab = newton (&det_onde_fonction,&det_onde_premiere,&det_onde_seconde,dop/2);
		ret.ampl = det_onde_fonction(etab);
		ret.speed = 2*M_PI/p;
		ret.etab = etab*ret.speed;
	}
	else
	{
		ret=det_onde_periode_fourier(p);
	}
	return (ret);
}

struct tonde det_onde_periode_fourier (LDOUBLE p)
{
	LDOUBLE gta,gtb,gtc,res;
	MINT i;
	LDOUBLE result;
	const LDOUBLE hinc = 3600;
	complex<float> w = complex<float>(0,2*M_PI/p);
	complex<float> resu=complex<float>(0,0);
	complex<float> ancexp=complex<float>(1,0);
	complex<float> expmul=exp(complex<float>(hinc,0)*w);
	complex<float> anccos[NUMCONST],mulcos[NUMCONST];
	for (i=0;i<NumOnde;i++)
	{
		anccos[i]=exp(complex<float>(0,-ondes[i].etab));
		mulcos[i]=exp(complex<float>(0,ondes[i].speed*hinc));
	} 
	struct tonde ret;
	LLONG n=0;
	LDOUBLE h;
	for (h=0;h<(LDOUBLE)(NUMJOURS-2)*24.*3600.;h+=hinc)
	{
		result = Z0;
		for (i=0;i<NumOnde;i++)
		{
			result+=ondes[i].ampl*real(anccos[i]);
			anccos[i]*=mulcos[i];
		}
		gta = tab[(LLONG)(h/PRECISION)];
		gtb = tab[(LLONG)(h/PRECISION)+1];
		gtc = h-((LLONG)(h/PRECISION))*PRECISION;
		res = ((gta*(PRECISION-gtc)+gtb*gtc)/PRECISION);
		resu+=complex<float>(((res/100)-result),0)*ancexp;
		ancexp*=expmul;
		//  resu+=rget_tide(h)*exp((complex<float>)w*h);
		n++;
		//  kbhit();
	}
	resu/=n;
	ret.speed = 2*M_PI/p;
	ret.ampl = 2*sqrt (norm(resu));
	ret.etab = arg(resu);
	return (ret);
}

struct tonde det_precis_func (DOUBLE p, DOUBLE pas, DOUBLE prec)
{
	struct tonde ret, maxret;
	DOUBLE per, tdep;
	DOUBLE ppas = pas/5;
	maxret.speed = 1;
	maxret.etab = 0;
	maxret.ampl = -100;
	for (per=p-1*pas;per<p+1*pas;per+=ppas)
	{
		if (((ret=det_onde_periode(per)).ampl) >= (maxret.ampl)) {maxret = ret;}
		kbhit (); // Pour Win95
	}
	if (ppas > prec) {maxret = det_precis_func (2*M_PI/maxret.speed,ppas,prec);}
	return (maxret);
}

struct tonde det_max_periode (void)
{
	FILE *fout;
	CHAR temp[100];
	MINT i;
	float far *RealIn,*RealOut,*CplxOut;
	long numpts=1;	// Transforme NUMPOINTS en puissance de 2
	float mmax,rmax,imax,pmax;
	struct tonde ret;
	while (numpts<NUMPOINTS) {numpts*=2;}
	// Allocation mmoire
	if (Ver_FFT) {MessageBox (NULL,"Allocation mmoire ...","FFT",MB_OK);}
	if ((RealIn = (float far *)malloc(numpts*4)) == NULL) {  exit (20);}
	if ((RealOut = (float far *)malloc(numpts*4)) == NULL) { exit (21);}
	if ((CplxOut = (float far *)malloc(numpts*4)) == NULL) { exit (22);}
	// Remplissage de RealIn : Tab - tide
	for (i=0;i<NUMPOINTS;i++) { RealIn [i]=tab[i]/100-tide(3600.*i); }   //TABPREC=3600 !!!!
	for (i=NUMPOINTS;i<numpts;i++) { RealIn [i] = 0; }
	// FFT de RealIn
	if (Ver_FFT) {MessageBox (NULL,"FFT de RealIn ...","FFT",MB_OK);}
	fft_float (numpts,0,RealIn,NULL,RealOut,CplxOut);
	// Determination du max.
	if (Ver_FFT) {MessageBox (NULL,"Determination du max ...","FFT",MB_OK);}
	mmax = rmax = imax = pmax = 0;
	for (i=5;i<(numpts/2);i++)
	{
		if ((RealOut[i]*RealOut[i]+CplxOut[i]*CplxOut[i])>mmax)
		{
			rmax = RealOut[i];
			imax = CplxOut[i];
			mmax = rmax*rmax+imax*imax;
			pmax = i;	
		}
	}
	/*/ Enregistrement sous fichier
	MessageBox (NULL,"Enregistrement ...","FFT",MB_OK);
	fout=fopen("fftresu.txt","w");
	for (i=0;i<(numpts/2);i++) { fprMINTf (fout,"%f;%f;%f;%f\n",(float)i*(1./3600.),RealOut[i],CplxOut[i],RealIn[i]); }
	fclose (fout);
	*/// Desallocation mmoire
	free (RealIn);
	free (RealOut);
	free (CplxOut);
	// Retour
	/*/ Retour sans precision
	ret.ampl = 2*sqrt (mmax);
	ret.etab = arg (complex<float>(rmax,imax));
	ret.speed = (pmax!=0)?(2.*M_PI)*pmax*(1./(numpts*3600.)):0; // TABPREC = 3600
	
	  sprMINTf (temp,"L'onde suivante a te ajoute :\n - Priode : %Lf\n - Amplitude : %Lf\n - Etablissement : %Lf\n%s\n",(2.*M_PI/ret.speed)/3600,ret.ampl,ret.etab,ret.name);
	  MessageBox (NULL, temp, "Slection d'une onde",MB_OK+MB_APPLMODAL+MB_ICONINFORMATION);
	*/
	ret = det_precis_func ((double)3600.*numpts/pmax,(double)numpts*3600./(pmax*pmax),0.00001);
	return (ret);
}